from enum import Enum

from .color import Color

__all__ = ["ThemeTypes", "SizeTypes", "NamedColor", "KmarkdownColors"]


class ThemeTypes(Enum):
    PRIMARY = "primary"
    SUCCESS = "success"
    DANGER = "danger"
    WARNING = "warning"
    INFO = "info"
    SECONDARY = "secondary"
    NONE = "none"


class SizeTypes(Enum):
    XS = "xs"
    SM = "sm"
    MD = "md"
    LG = "lg"


class KmarkdownColors(Enum):
    DANGER = "danger"
    INFO = "info"
    PINK = "pink"
    PRIMARY = "primary"
    SECONDARY = "secondary"
    SUCCESS = "success"
    WARNING = "warning"
    NONE = "none"


class NamedColor(Enum):
    BLACK = Color(0, 0, 0)
    DARK_BLUE = Color(0, 0, 170)
    DARK_GREEN = Color(0, 170, 0)
    DARK_AQUA = Color(0, 170, 170)
    DARK_RED = Color(170, 0, 0)
    DARK_PURPLE = Color(170, 0, 170)
    GOLD = Color(255, 170, 0)
    GRAY = Color(170, 170, 170)
    DARK_GRAY = Color(85, 85, 85)
    BLUE = Color(85, 85, 255)
    GREEN = Color(85, 255, 85)
    AQUA = Color(85, 255, 255)
    RED = Color(255, 85, 85)
    YELLOW = Color(255, 255, 85)
    WHITE = Color(255, 255, 255)



import json
from collections.abc import Sequence
from typing import List, Union, TypeVar, Iterator, Optional

from .color import Color
from .modules import _Module
from .types import SizeTypes, NamedColor, ThemeTypes

__all__ = ["Card", "CardMessage"]

_T_co = TypeVar("_T_co", covariant=True)


class Card(Sequence):
    """
    构建卡片
    """

    type: str = "card"
    theme: str
    size: str
    color: Optional[str]
    modules: List[_Module]

    def __init__(
        self,
        *modules: _Module,
        theme: Union[str, ThemeTypes] = ThemeTypes.PRIMARY,
        size: Union[str, SizeTypes] = SizeTypes.LG,
        color: Union[Color, NamedColor, str, None] = None,
    ) -> None:
        """
        构建卡片

        :param modules: 卡片模块列表
        :param theme: 卡片主题
        :param size: 目前只支持sm与lg。 lg仅在PC端有效, 在移动端不管填什么，均为sm。
        :param color: 卡片颜色 ex: #55ffff or NamedColor.XXX
        """
        self.modules = list(modules)
        self.theme = theme if isinstance(theme, str) else theme.value
        self.size = size if isinstance(size, str) else size.value
        if color is None:
            self.color = None
        elif isinstance(color, str):
            self.color = color
        elif isinstance(color, Color):
            self.color = color.__str__()
        elif isinstance(color, NamedColor):
            self.color = color.value.__str__()
        else:
            raise ValueError("incorrect color value: " + self.color)

    def __getitem__(self, item: int) -> _Module:
        return self.modules[item]

    def __setitem__(self, key: int, value: _Module):
        self.modules[key] = value

    def __len__(self):
        return len(self.modules)

    def __iter__(self) -> Iterator[_T_co]:
        return self.modules.__iter__()

    def __contains__(self, value: object) -> bool:
        return self.modules.__contains__(value)

    def __reversed__(self) -> Iterator[_T_co]:
        return self.modules.__reversed__()

    def __repr__(self):
        return "Card(" + ", ".join([module.__repr__() for module in self.modules]) + ")"

    def count(self, value: _Module) -> int:
        return self.modules.count(value)

    def index(self, value: _Module, start: int = ..., stop: int = ...) -> int:
        return self.modules.index(value, start, stop)

    def append(self, module: _Module):
        self.modules.append(module)

    def build(self) -> dict:
        """
        :return: 构造后卡片
        """
        ret = {"type": self.type, "theme": self.theme, "size": self.size, "modules": []}
        if self.color is not None:
            ret["color"] = self.color
        for i in self.modules:
            ret["modules"].append(i.build())
        return ret

    def build_to_json(self) -> str:
        return json.dumps(self.build(), indent=4, ensure_ascii=False)

    def clear(self) -> "Card":
        self.modules.clear()
        return self

    def set_theme(self, theme: Union[str, ThemeTypes]) -> "Card":
        self.theme = theme if isinstance(theme, str) else theme.value
        return self

    def set_size(self, size: Union[str, SizeTypes]) -> "Card":
        self.size = size if isinstance(size, str) else size.value
        return self

    def set_color(self, color: Union[str, Color, NamedColor, None]) -> "Card":
        if color is None:
            self.color = None
        elif isinstance(color, str):
            self.color = color
        elif isinstance(color, Color):
            self.color = color.__str__()
        elif isinstance(color, NamedColor):
            self.color = color.value.__str__()
        else:
            raise ValueError("incorrect color value: " + self.color)
        return self


class CardMessage(Sequence):
    card_list: List[Card]
    """
    构建卡片列表
    """

    def __init__(self, *card: Card) -> None:
        """构建卡片列表,"""
        self.card_list = list(card)

    def __getitem__(self, item: int) -> Card:
        return self.card_list[item]

    def __setitem__(self, key: int, value: Card):
        self.card_list[key] = value

    def __len__(self):
        return len(self.card_list)

    def __iter__(self) -> Iterator[_T_co]:
        return self.card_list.__iter__()

    def __contains__(self, value: object) -> bool:
        return self.card_list.__contains__(value)

    def __reversed__(self) -> Iterator[_T_co]:
        return self.card_list.__reversed__()

    def __repr__(self):
        return (
            "CardMessage("
            + ", ".join([card.__repr__() for card in self.card_list])
            + ")"
        )

    def index(self, value: Card, start: int = ..., stop: int = ...) -> int:
        return self.card_list.index(value, start, stop)

    def count(self, value: Card) -> int:
        return self.card_list.count(value)

    def append(self, card: Card):
        self.card_list.append(card)

    def build(self):
        return [card.build() for card in self.card_list]

    def build_to_json(self) -> str:
        return json.dumps(self.build(), indent=4, ensure_ascii=False)



from .accessory import Button, Image, Kmarkdown, Paragraph, PlainText
from .builder import (
    ActionGroupBuilder,
    CardBuilder,
    CardMessageBuilder,
    ContainerBuilder,
    ContextBuilder,
    ImageGroupBuilder,
)
from .card import Card, CardMessage
from .color import Color
from .modules import (
    ActionGroup,
    Audio,
    Container,
    Context,
    Countdown,
    Divider,
    File,
    Header,
    ImageGroup,
    Invite,
    Section,
    Video,
)
from .types import KmarkdownColors, NamedColor, SizeTypes, ThemeTypes



from abc import ABC, abstractmethod
from typing import List, Optional, Union

from .accessory import (
    Button,
    Image,
    Paragraph,
    PlainText,
    _BaseAccessory,
    _BaseNonText,
    _BaseText,
)
from .card import Card, CardMessage
from .modules import (
    ActionGroup,
    Audio,
    Container,
    Context,
    Countdown,
    Divider,
    File,
    Header,
    ImageGroup,
    Invite,
    Section,
    Video,
)

__all__ = [
    "CardMessageBuilder",
    "CardBuilder",
    "ImageGroupBuilder",
    "ContainerBuilder",
    "ContextBuilder",
    "ActionGroupBuilder",
]


class AbstractBuilder(ABC):
    @abstractmethod
    def build(self):
        ...


class CardMessageBuilder(AbstractBuilder):
    def __init__(self) -> None:
        self.__card_message = CardMessage()

    def card(self, card: Card):
        """添加一个卡片到卡片消息中"""
        self.__card_message.append(card)
        return self

    def build(self) -> CardMessage:
        """构造为 CardMessage"""
        return self.__card_message


class CardBuilder(AbstractBuilder):
    def __init__(self) -> None:
        self._card = Card()

    def header(self, text: Union[str, PlainText] = ""):
        """
        为卡片添加一个 header

        :param text: 标题内容
        """
        self._card.append(Header(text))
        return self

    def section(
        self,
        text: Union[_BaseText, Paragraph],
        *,
        mode: str = "right",
        accessory: _BaseNonText = None,
    ):
        """
        为卡片添加一个 section

        :param text: 文本元素或结构体
        :param mode: accessory在左侧还是在右侧 只能为 left|right
        :param accessory: 非文本元素
        """
        self._card.append(Section(text, mode=mode, accessory=accessory))
        return self

    def image_group(self, image_group: ImageGroup):
        """
        为卡片添加一个 image_group

        :param image_group: 要添加的 ImageGroup
        """
        self._card.append(image_group)
        return self

    def container(self, container: Container):
        """
        为卡片添加一个 container

        :param container: 要添加的 Container
        """
        self._card.append(container)
        return self

    def action_group(self, action_group: ActionGroup):
        """
        为卡片添加一个 action_group

        :param action_group: 要添加的 ActionGroup
        """
        self._card.append(action_group)
        return self

    def context(self, context: Context):
        """
        为卡片添加一个 context

        :param context: 要添加的 Context
        """
        self._card.append(context)
        return self

    def divider(self):
        """为卡片添加一个 divider"""
        self._card.append(Divider())
        return self

    def day_countdown(self, end_time: str):
        """
        为卡片添加一个 天-倒计时

        :param end_time: 结束时间 ex: 2022-05-05 08:00:00
        """
        self._card.append(Countdown.new_day_countdown(end_time))
        return self

    def hour_countdown(self, end_time: str):
        """
        为卡片添加一个 小时-倒计时

        :param end_time: 结束时间 ex: 2022-05-05 08:00:00
        """
        self._card.append(Countdown.new_hour_countdown(end_time))
        return self

    def second_countdown(self, end_time: str, start_time: Optional[str] = None):
        """
        为卡片添加一个 秒-倒计时

        :param end_time: 结束时间 ex: 2022-05-05 08:00:00
        :param start_time: 开始时间 ex: 2022-05-05 08:00:00 留空则为当前时间
        """
        self._card.append(Countdown.new_second_countdown(end_time, start_time))
        return self

    def invite(self, code: str):
        """
        为卡片添加一个 invite

        :param code: 邀请码
        """
        self._card.append(Invite(code))
        return self

    def file(self, src: str, title: str):
        """
        为卡片添加一个 file

        :param src: 文件链接
        :param title: 文件标题
        """
        self._card.append(File(src, title))
        return self

    def video(self, src: str, title: str):
        """
        为卡片添加一个 video

        :param src: 视频链接
        :param title: 视频标题
        """
        self._card.append(Video(src, title))
        return self

    def audio(self, src: str, title: str, cover: Optional[str] = None):
        """
        为卡片添加一个 audio

        :param src: 音乐链接
        :param title: 音乐标题
        :param cover: 音乐封面
        """
        self._card.append(Audio(src, title, cover))
        return self

    def build(self) -> Card:
        return self._card

    @property
    def card(self):
        """非必要不访问"""
        return self._card


class MultiBuilder(AbstractBuilder, ABC):
    elements: List[_BaseAccessory]

    def __init__(self) -> None:
        self.elements = []

    @abstractmethod
    def add(self, accessory: _BaseAccessory):
        ...

    @abstractmethod
    def build(self):
        ...


class ImageGroupBuilder(MultiBuilder):
    def add(self, accessory: Image):
        self.elements.append(accessory)
        return self

    def build(self) -> ImageGroup:
        return ImageGroup(*self.elements)


class ContainerBuilder(MultiBuilder):
    def add(self, accessory: Image):
        self.elements.append(accessory)
        return self

    def build(self) -> Container:
        return Container(*self.elements)


class ActionGroupBuilder(MultiBuilder):
    def add(self, accessory: Button):
        self.elements.append(accessory)
        return self

    def build(self) -> ActionGroup:
        return ActionGroup(*self.elements)


class ContextBuilder(MultiBuilder):
    def add(self, accessory: _BaseAccessory):
        self.elements.append(accessory)
        return self

    def build(self) -> Context:
        return Context(*self.elements)



import json
from abc import ABC, abstractmethod
from typing import List, Union

__all__ = [
    "PlainText",
    "Kmarkdown",
    "Paragraph",
    "Image",
    "Button",
    "_BaseAccessory",
    "_BaseText",
    "_BaseNonText",
]

from .types import KmarkdownColors, SizeTypes, ThemeTypes


class _BaseAccessory(ABC):
    """
    元素基类
    """

    @abstractmethod
    def build(self) -> dict:
        """
        :return: 构造后元素
        """

    def build_to_json(self) -> str:
        return json.dumps(self.build(), indent=4, ensure_ascii=False)

    @abstractmethod
    def __repr__(self):
        ...


class _BaseText(_BaseAccessory, ABC):
    """
    文字类元素基类
    """

    content: str


class _BaseNonText(_BaseAccessory, ABC):
    """
    非文字类元素基类
    """


class PlainText(_BaseText):
    """
    构造纯文本元素
    """

    emoji: bool

    def __init__(self, content: str = "", emoji=True) -> None:
        """
        构造纯文本

        :param content: 文本内容
        :param emoji: 默认为 true。如果为 true,会把 emoji 的 shortcut 转为 emoji
        """
        self.type = "plain-text"
        self.content = content
        self.emoji = emoji

    def build(self) -> dict:
        return {"type": self.type, "content": self.content}

    def __repr__(self):
        return f"PlainText(content='{self.content}', emoji={self.emoji})"


class Kmarkdown(_BaseText):
    """
    构造kmarkdown文本元素
    """

    def __init__(self, content: str = "") -> None:
        """
        构造kmarkdown文本

        :param content: kmarkdown文本
        """
        self.type = "kmarkdown"
        self.content = content

    @classmethod
    def bold(cls, content: str = ""):
        """构造加粗文字"""
        return cls(f"**{content}**")

    @classmethod
    def italic(cls, content: str = ""):
        """构造斜体文字"""
        return cls(f"*{content}*")

    @classmethod
    def bold_italic(cls, content: str = ""):
        """构造加粗斜体文字"""
        return cls(f"***{content}***")

    @classmethod
    def strikethrough(cls, content: str = ""):
        """构造删除线文字"""
        return cls(f"~~{content}~~")

    @classmethod
    def link(cls, text: str, link: str):
        """构造超链接文字"""
        return cls(f"[{text}]({link})")

    @classmethod
    def divider(cls):
        return cls("---")

    @classmethod
    def quote(cls, content: str = ""):
        return cls(f"> {content}")

    @classmethod
    def underline(cls, content: str = ""):
        return cls(f"(ins){content}(ins)")

    @classmethod
    def spoiler(cls, content: str = ""):
        return cls(f"(spl){content}(spl)")

    @classmethod
    def at_channel(cls, channel_id: str):
        return cls(f"(chn){channel_id}(chn)")

    @classmethod
    def at_user(cls, user_id: str):
        return cls(f"(met){user_id}(met)")

    @classmethod
    def at_role(cls, role_id: str):
        return cls(f"(rol){role_id}(rol)")

    @classmethod
    def inline_code(cls, code: str):
        return cls(f"`{code}`")

    @classmethod
    def code_block(cls, code: str, language: str = ""):
        return cls(f"```{language}\n{code}\n```")

    @classmethod
    def color(
        cls, content: str, color: Union[str, KmarkdownColors] = KmarkdownColors.NONE
    ):
        return cls(
            f"(font){content}(font)[{color if isinstance(color, str) else color.value}]"
        )

    def build(self) -> dict:
        return {"type": self.type, "content": self.content}

    def __repr__(self):
        return f"Kmarkdown(content='{self.content}')"

    def __add__(self, other: "Kmarkdown") -> "Kmarkdown":
        self.content += other.content
        return self


class Paragraph(_BaseAccessory):
    """
    构造多列文本元素
    """

    cols: int
    fields: List[_BaseText]

    def __init__(self, cols: int, fields: List[_BaseText]) -> None:
        """
        构造多列文本

        :param cols: 列数 只能为 1-3
        :param fields: 文本组件列表
        """
        if not (1 <= cols <= 3):
            raise Exception("文本列数不为 1-3")
        if len(fields) != cols:
            raise Exception("文本列数与列表不符")
        for i in fields:
            if isinstance(i, Paragraph):
                raise Exception("文本组件不能为paragraph")
        self.type = "paragraph"
        self.cols = cols
        self.fields = fields

    def build(self) -> dict:
        ret = {"type": self.type, "cols": self.cols, "fields": []}
        for i in self.fields:
            ret["fields"].append(i.build())
        return ret

    def __repr__(self):
        return f'Paragraph(cols={self.cols}, fields=[{", ".join([text.__repr__() for text in self.fields])}])'


class Image(_BaseNonText):
    """
    显示图片元素
    """

    src: str
    alt: str
    size: str
    circle: bool

    def __init__(
        self,
        src: str,
        size: Union[str, SizeTypes] = "lg",
        alt: str = "",
        circle: bool = False,
    ) -> None:
        """
        显示图片元素

        :param src: 图片地址
        :param size: 图片大小样式 只能为 sm 或 lg
        :param alt: 不知道干嘛用的
        :param circle: 显示圆形图片，在文本+图片时有效
        """
        self.type = "image"
        self.src = src
        self.alt = alt
        self.size = size.value if isinstance(size, SizeTypes) else size
        self.circle = circle

    def build(self) -> dict:
        return {
            "type": self.type,
            "src": self.src,
            "alt": self.alt,
            "size": self.size,
            "circle": self.circle,
        }

    def __repr__(self):
        return f"Image(src='{self.src}', size='{self.size}', alt='{self.alt}', circle={self.circle})"


class Button(_BaseNonText):
    theme: str
    value: str
    click: str
    text: _BaseText

    def __init__(
        self,
        text: _BaseText,
        theme: Union[str, ThemeTypes] = "primary",
        value: str = "",
        click: str = "",
    ) -> None:
        self.type = "button"
        self.text = text
        self.theme = theme if isinstance(theme, str) else theme.value
        self.value = value
        self.click = click

    def build(self) -> dict:
        return {
            "type": self.type,
            "theme": self.theme,
            "value": self.value,
            "click": self.click,
            "text": self.text.build(),
        }

    def __repr__(self):
        return f"Button(text={self.text.__repr__()}, theme='{self.theme}', value='{self.value}', click='{self.click}')"
